import {
  json
} from "@remix-run/server-runtime";
import {
  Strategy
} from "remix-auth";
import {
  verifyRegistrationResponse,
  verifyAuthenticationResponse
} from "@simplewebauthn/server";
class WebAuthnStrategy extends Strategy {
  constructor(options, verify) {
    super(verify);
    this.name = "webauthn";
    this.rpName = options.rpName;
    this.rpID = options.rpID;
    this.origin = options.origin;
    this.getUserAuthenticators = options.getUserAuthenticators;
    this.getUserDetails = options.getUserDetails;
    this.getUserByUsername = options.getUserByUsername;
    this.getAuthenticatorById = options.getAuthenticatorById;
  }
  async authenticate(request, sessionStorage, options) {
    var _a;
    let session = await sessionStorage.getSession(
      request.headers.get("Cookie")
    );
    try {
      let user = (_a = session.get(options.sessionKey)) != null ? _a : null;
      if (user && request.method === "POST") {
        return this.success(user, request, sessionStorage, options);
      }
      const rp = {
        name: typeof this.rpName === "function" ? await this.rpName(request) : this.rpName,
        id: typeof this.rpID === "function" ? await this.rpID(request) : this.rpID
      };
      const origin = typeof this.origin === "function" ? await this.origin(request) : this.origin;
      if (request.method === "GET") {
        let authenticators = [];
        let userDetails = null;
        let usernameAvailable = null;
        if (!user) {
          const username2 = new URL(request.url).searchParams.get("username");
          if (username2) {
            usernameAvailable = true;
            user = await this.getUserByUsername(username2 || "");
          }
        }
        if (user) {
          authenticators = await this.getUserAuthenticators(user);
          userDetails = await this.getUserDetails(user);
          usernameAvailable = false;
        }
        const crypto = await import("tiny-webcrypto");
        const options2 = {
          usernameAvailable,
          rp,
          user: userDetails ? { displayName: userDetails.username, ...userDetails } : null,
          challenge: Buffer.from(
            crypto.default.getRandomValues(new Uint8Array(32))
          ).toString("base64url"),
          authenticators: authenticators.map(
            ({ credentialID, transports }) => ({
              id: credentialID,
              type: "public-key",
              transports
            })
          )
        };
        session.set("challenge", options2.challenge);
        throw json(options2, {
          status: 200,
          headers: {
            "Set-Cookie": await sessionStorage.commitSession(session),
            "Cache-Control": "no-store"
          }
        });
      }
      if (request.method !== "POST")
        throw new Error(
          "Only use the WebAuthn authenticate with POST or GET requests."
        );
      const expectedChallenge = session.get("challenge");
      const formData = await request.formData();
      let data;
      try {
        const responseData = formData.get("response");
        if (typeof responseData !== "string")
          throw new Error("Error");
        data = JSON.parse(responseData);
      } catch {
        throw new Error("Invalid passkey response JSON.");
      }
      const type = formData.get("type");
      let username = formData.get("username");
      if (typeof username !== "string")
        username = null;
      if (type === "registration") {
        if (!username)
          throw new Error("Username is a required form value.");
        const verification = await verifyRegistrationResponse({
          response: data,
          expectedChallenge,
          expectedOrigin: origin,
          expectedRPID: rp.id
        });
        if (verification.verified && verification.registrationInfo) {
          const {
            credentialPublicKey,
            credentialID,
            counter,
            credentialBackedUp,
            credentialDeviceType
          } = verification.registrationInfo;
          const newAuthenticator = {
            credentialID: Buffer.from(credentialID).toString("base64url"),
            credentialPublicKey: Buffer.from(credentialPublicKey).toString("base64url"),
            counter,
            credentialBackedUp: credentialBackedUp ? 1 : 0,
            credentialDeviceType,
            transports: ""
          };
          user = await this.verify({
            authenticator: newAuthenticator,
            type: "registration",
            username
          });
        } else {
          throw new Error("Passkey verification failed.");
        }
      } else if (type === "authentication") {
        const authenticationData = data;
        const authenticator = await this.getAuthenticatorById(
          authenticationData.id
        );
        if (!authenticator)
          throw new Error("Passkey not found.");
        const verification = await verifyAuthenticationResponse({
          response: authenticationData,
          expectedChallenge,
          expectedOrigin: origin,
          expectedRPID: rp.id,
          authenticator: {
            ...authenticator,
            credentialPublicKey: Buffer.from(
              authenticator.credentialPublicKey,
              "base64url"
            ),
            credentialID: Buffer.from(authenticator.credentialID, "base64url"),
            transports: authenticator.transports.split(
              ","
            )
          }
        });
        if (!verification.verified)
          throw new Error("Passkey verification failed.");
        user = await this.verify({
          authenticator,
          type: "authentication",
          username
        });
      } else {
        throw new Error("Invalid verification type.");
      }
      return this.success(user, request, sessionStorage, options);
    } catch (error) {
      if (error instanceof Response)
        throw error;
      if (error instanceof Error) {
        return await this.failure(
          error.message,
          request,
          sessionStorage,
          options,
          error
        );
      }
      if (typeof error === "string") {
        return await this.failure(
          error,
          request,
          sessionStorage,
          options,
          new Error(error)
        );
      }
      return await this.failure(
        "Unknown error",
        request,
        sessionStorage,
        options,
        new Error(JSON.stringify(error, null, 2))
      );
    }
  }
}
export {
  WebAuthnStrategy
};
